from allennlp_models.vision.predictors.vilbert_ir import VilbertImageRetrievalPredictor
from allennlp_models.vision.predictors.vilbert_vqa import VilbertVqaPredictor
from allennlp_models.vision.predictors.visual_entailment import VisualEntailmentPredictor
