from allennlp_models.pair_classification.dataset_readers.quora_paraphrase import (
    QuoraParaphraseDatasetReader,
)
from allennlp_models.pair_classification.dataset_readers.snli import SnliReader
from allennlp_models.pair_classification.dataset_readers.transformer_superglue_rte import (
    TransformerSuperGlueRteReader,
)
