from allennlp_models.vision.models.heads.nlvr2_head import Nlvr2Head
from allennlp_models.vision.models.heads.vqa_head import VqaHead
from allennlp_models.vision.models.heads.visual_entailment_head import VisualEntailmentHead
